module Main where

import           Control.Arrow (second)
import           Control.Parallel.Strategies (parMap, rdeepseq)
import qualified Data.ByteString.Char8 as BC
import           Data.Foldable (toList)
import           Data.List (intercalate, transpose)
import qualified Data.Set as Set
import           System.Environment (getArgs, getProgName)
import           System.Exit (exitFailure)
import           System.IO (hPutStrLn, stderr)

import           Combine
import           Interpreter
import           Predictor (showSampleLabels)
import           Types


usageInfo :: String -> String
usageInfo progName = intercalate "\n"
  [ "Usage: " ++ progName ++ " <seq> <eval> <mean> <maxRank> "
    ++ "[<file1> <weight1> [... [<file_n> <weight_n>]]]"
  , "Combine and output predictions using a voting scheme."
  , ""
  , "The predicted usefulness of a given premise is computed using a "
  , "weighted average of the usefulnesses from the combined files."
  , ""
  , "The possible averages are:"
  , "  Arithmetic, Geometric, Quadratic, Maximum, or Harmonic (typically best)"
  , ""
  , "A prediction file specifies maximally n predictions for each theorem."
  , "The votes for the first n are their number in the sequence, and"
  , "for the rest, we consider their vote to be the given number `maxRank'."
  , ""
  , "Example usage:"
  , "  " ++ progName ++ " seq eval Harmonic 1024 file1 0.3 file2 0.4 file3 0.3"
  ]

showUsage :: IO ()
showUsage = do
  progName <- getProgName
  hPutStrLn stderr $ usageInfo progName

-- | Interpret a sequence of path-weight pairs, such as "file1 0.3 file2 0.4".
parsePreds :: [String] -> [(FilePath, Weight)]
parsePreds (path : weight : xs) = (path, read weight) : parsePreds xs
parsePreds [_] = error "prediction weight not specified"
parsePreds [ ] = []


main :: IO ()
main = do
  args <- getArgs
  (seqFP, evalFP, mean, maxRank, predictions) <- case args of
    (sF : eF : m : mR : p) -> return (sF, eF, read m, read mR, parsePreds p)
    _ -> showUsage >> exitFailure

  let (predFPs, predWeights) = unzip predictions
      combineF = take maxRank . combine mean maxRank predWeights

  lblMap        <- loadSequenceFile seqFP
  eval          <- Set.fromList `fmap` loadEvaluationFile lblMap evalFP
  fileLineLabel <- mapM (loadLabelsFile lblMap) predFPs

  let lineFileLabel = zip (map SampleID [0 ..]) $ transpose $ map toList fileLineLabel
      evaluated = [(i, l) | (i, l) <- lineFileLabel, i `Set.member` eval]
      combined = parMap rdeepseq (second combineF) evaluated
  mapM_ (BC.putStrLn . uncurry (showSampleLabels lblMap)) combined
