#!/usr/bin/env runhaskell

import Data.List (intercalate)
import Data.Maybe (fromJust)
import Development.Shake
import Development.Shake.FilePath


-- -----------------------------------------------------------------------------
-- Experiment definition

-- | Experiment consists of a name and a list of files.
type Experiment = (String, [FilePath])

-- | Obtain the name of an experiment.
experimentName :: Experiment -> String
experimentName = fst

-- | Files of an experiment, with experiment name added to path.
experimentFiles :: Experiment -> [FilePath]
experimentFiles (name, files) = map ((</>) name) files

-- Given a list of experiments, retrieve the experiment with the given name.
experimentByName :: [Experiment] -> String -> Experiment
experimentByName exs name = (name, fromJust $ lookup name exs)


-- -----------------------------------------------------------------------------
-- Experiments

-- | Combine predictor options.
opts :: [String] -> String
opts = intercalate ":"

-- | Number of trees versus sample frequency.
exnTsF :: Experiment
exnTsF = ("nTsF", map opts
  [ ["rforest", "nTrees=" ++ show nT, "sampleFreq=" ++ show sF]
  | nT <- [2, 4, 6, 8, 12, 16, 24, 32 :: Int]
  , sF <- [1, 2, 4, 6, 8, 12, 16, 24, 32 :: Int]
  ])

-- | Prediction algorithms.
exPredAlg :: Experiment
exPredAlg = ("predAlg",
  [ "knn"
  , "rforest:predAlg=Naive"
  , "rforest:predAlg=KNN:tfidf=False"
  , "rforest:predAlg=KNN:tfidf=True"
  ])

-- | Single path querying.
exSPQ :: Experiment
exSPQ = ("SPQ",
  [ "rforest:depthWeightAlg=Constant:minDepthWeight=0"
  ])

-- | Random features.
-- TODO: Does currently not work!
exRandomFeats :: Experiment
exRandomFeats = ("randomFeats", map opts
  [ ["rforest", "randomFeats=" ++ show rF]
  | rF <- [2, 4, 6, 8, 12, 16 :: Int]
  ])

-- | Gini features.
exGiniFeats :: Experiment
exGiniFeats = ("giniFeats", map opts
  [ ["rforest", "giniFeats=" ++ show gF]
  | gF <- [2, 4, 6, 8, 12, 16 :: Int]
  ])

-- | Bounded tree depths.
exTreeDepth :: Experiment
exTreeDepth = ("treeDepth", map opts
  [ ["rforest", "maxTreeDepth=" ++ show mTD]
  | mTD <- [2, 4, 6, 8, 12, 16, 24, 32 :: Int]
  ])

-- | Strategies for assigning samples to trees.
exSampleStrat :: Experiment
exSampleStrat = ("sampleStrat", map opts
  [ ["rforest", "sampleStrat=" ++ sS]
  | sS <- ["TreesDrawSamples", "SamplesDrawTrees"]
  ])

-- | Path score algorithms.
exPathScore :: Experiment
exPathScore = ("pathScore", map opts
  [ ["rforest", "depthWeightAlg=" ++ dWA]
  | dWA <- ["Ascending", "Descending", "Inverse", "Constant"]
  ])

-- | List of all experiments.
experiments :: [Experiment]
experiments =
  [ exPredAlg
  , exnTsF
  , exSPQ
  --, exRandomFeats
  , exGiniFeats
  , exTreeDepth
  , exSampleStrat
  , exPathScore
  ]


-- -----------------------------------------------------------------------------
-- Test sets

data TestSet = TestSet
  { predictCmd :: FilePath
  , statisticsCmd :: FilePath
  , evalFile :: FilePath
  , experimentDir :: FilePath
  , statisticsDir :: FilePath
  }

-- | Create a test set from a test set name.
generateTestSet :: String -> TestSet
generateTestSet name = TestSet
  { predictCmd = "./predict" ++ name ++ ".sh"
  , statisticsCmd = "./statistics" ++ name ++ ".sh"
  , evalFile = name </> "eval"
  , experimentDir = "experiments-" ++ name
  , statisticsDir = "statistics-" ++ name
  }

mizarSet :: TestSet
mizarSet = generateTestSet "mizar"

isabelleSet :: TestSet
isabelleSet = generateTestSet "isabelle"


-- -----------------------------------------------------------------------------
-- Build

-- | Run prediction. The name of the prediction file is equivalent to the
-- options passed to the predictor.
predict :: TestSet -> FilePath -> Action ()
predict (TestSet {predictCmd = pC, evalFile = eF}) out =
  cmd Shell pC "-e" eF "-p" [takeFileName out] ">" [out]

-- | Generate prediction statistics.
-- If the predictions do not exist yet, they will be generated here.
statistics :: TestSet -> FilePath -> Action ()
statistics (TestSet {statisticsCmd = sC, experimentDir = eD}) out = do
  need exFiles
  cmd Shell sC exFiles ">" [out]
  where ex = experimentByName experiments (dropDirectory1 out)
        exFiles = map ((</>) eD) $ experimentFiles ex

-- Final files produced by given test set experiments.
filesTestSet :: TestSet -> [Experiment] -> [FilePath]
filesTestSet set = map ((</>) (statisticsDir set) . experimentName)

-- | Rules for building a test set.
rulesTestSet :: TestSet -> Rules ()
rulesTestSet set = do
  (experimentDir set ++ "//*") %> predict set
  (statisticsDir set ++ "//*") %> statistics set

-- | Run experiments.
main :: IO ()
main = shakeArgs shakeOptions $ do
  mapM_ rulesTestSet [mizarSet, isabelleSet]
  want $ filesTestSet mizarSet experiments
