{-# LANGUAGE FlexibleContexts
           , OverloadedStrings
           , TypeFamilies #-}

module Predictor where

import           Control.Monad.Reader (Reader, ReaderT, ask, runReader, runReaderT)
import           Control.Monad.Reader.Class (MonadReader)
import           Control.Monad (liftM)
import           Data.ByteString (ByteString)
import qualified Data.ByteString.Char8 as BC

import qualified IntBiMap as IBM
import           Types


-- -----------------------------------------------------------------------------
-- Types

type PredictorData p = (LearnData, PredictorCfg p)
type PredictorM p = Reader  (PredictorData p)
type PredictorT p = ReaderT (PredictorData p)

class Predictor p where
  data PredictorCfg p :: *

  learn      :: p -> Samples  -> (PredictorM p) p
  predict    :: p -> Features -> (PredictorM p) Labels
  saveState  :: p -> FilePath -> (PredictorT p) IO ()
  loadState  :: p -> FilePath -> (PredictorT p) IO p


-- -----------------------------------------------------------------------------
-- Reader wrappers

runPredictorT :: PredictorT p m a -> LearnData -> PredictorCfg p -> m a
runPredictorT p l c = runReaderT p (l, c)

-- also present in Control.Monad.Trans.InterleavableIO
promoteReader :: MonadReader s m => Reader s a -> m a
promoteReader act = liftM (runReader act) ask

promotePredictor
  :: MonadReader (PredictorData p) m => Reader (PredictorData p) a -> m a
promotePredictor = promoteReader


-- -----------------------------------------------------------------------------
-- Reader data retrieval

askLData :: MonadReader (PredictorData p) m => m LearnData
askLData = liftM fst ask

asksLData :: MonadReader (PredictorData p) m => (LearnData -> a) -> m a
asksLData f = liftM f askLData

askConfig :: (PredictorM p) (PredictorCfg p)
askConfig = liftM snd ask

asksConfig :: (PredictorCfg p -> a) -> (PredictorM p) a
asksConfig f = liftM f askConfig


-- -----------------------------------------------------------------------------
-- Printing

showSample :: LabelMap -> SampleID -> ByteString
showSample m (SampleID s) = IBM.findElement m s

showSampleM :: SampleID -> (PredictorM p) ByteString
showSampleM s = asksLData labelMap >>= \ lM -> return $ showSample lM s

showLabel :: LabelMap -> LabelID -> ByteString
showLabel = IBM.findElement

showLabelM :: LabelID -> (PredictorM p) ByteString
showLabelM l = asksLData labelMap >>= \ lM -> return $ showLabel lM l

showLabels :: LabelMap -> Labels -> ByteString
showLabels lM = BC.unwords . map (showLabel lM)

showLabelsM :: Labels -> (PredictorM p) ByteString
showLabelsM = liftM BC.unwords . mapM showLabelM

showSampleLabels :: LabelMap -> SampleID -> Labels -> ByteString
showSampleLabels lM s ls = BC.concat [showSample lM s, ":", showLabels lM ls]

showSampleLabelsM :: SampleID -> Labels -> (PredictorM p) ByteString
showSampleLabelsM s ls = asksLData labelMap >>= \ lM ->
  return $ showSampleLabels lM s ls

showFeatureM :: FeatureID -> (PredictorM p) ByteString
showFeatureM f = asksLData featureMap >>= \ fM -> return $ IBM.findElement fM f
