import           Control.Applicative (pure)
import           Control.Arrow (second)
import           Control.DeepSeq (NFData, rnf)
import           Control.Monad (liftM, when)
import           Control.Parallel.Strategies (parMap, rdeepseq)
import           Data.Foldable (toList)
import qualified Data.ByteString.Char8 as BC
import qualified Data.IntMap as IM
import qualified Data.IntSet as IS
import           Data.List (findIndices, intercalate, sortBy, transpose)
import           Data.Maybe (mapMaybe)
import           Data.Ord (comparing)
import           Data.Sequence (Seq)
import qualified Data.Sequence as Seq
import           System.Console.GetOpt
import           System.Environment (getArgs, getProgName)
import           System.Exit (exitFailure)
import           System.IO (hPutStrLn, stderr)
import           Text.Printf (printf)

import qualified IntBiMap as IBM
import qualified Interpreter as I
import           Table
import           Types


data Statistics = Statistics {
  auc       :: Double
, cover     :: Double
, rank      :: Double
, precision :: Int
, recall    :: Int
, coverI    :: [Double]
, recallI   :: [Int]
, lblRanks  :: [(LabelID, Maybe Int)]
} deriving Show

data FileStatistics = FileStatistics {
  aucF       :: Double
, coverF     :: Double
, rankF      :: Double
, precisionF :: Double
, recallF    :: Double
, coverIF    :: [Double]
, recallIF   :: [Int]
, lblRanksF  :: [(LabelID, [(LabelID, Maybe Int)])]
}

instance NFData FileStatistics where
  rnf (FileStatistics auc' cover' rank' prec' rec' covi' reci' lr') =
    rnf auc' `seq` rnf cover' `seq` rnf rank' `seq` rnf prec' `seq`
    rnf rec' `seq` rnf covi'  `seq` rnf reci' `seq` rnf lr'

calculateAUC :: Int -> Int -> [Int] -> Double
calculateAUC nLabels nPreds hitIndices =
  if nLabels == 0 || neg == 0 then 1
  else fromIntegral (sum areas) / fromIntegral (nLabels * neg)
  where neg = nPreds - length hitIndices
        areas = zipWith3 (\ l h n -> (h - l) * n) (init hIs) (tail hIs) [1 ..]
        hIs = hitIndices ++ [nPreds]

calculateRank :: Int -> Int -> [Int] -> Double
calculateRank nDeps nPreds hitIndices =
  fromIntegral (sum hitIndices + missed * nPreds) / fromIntegral nDeps
  where missed = nDeps - length hitIndices

calculateRecall :: Int -> Int -> [Int] -> Int
calculateRecall nDeps nPreds hitIndices
  | length hitIndices == nDeps = 1 + last hitIndices
  | otherwise                  = 1 + nPreds

calculateCoverI :: Int -> [Int] -> [Int] -> [Double]
calculateCoverI nLabels hitIndices =
  map (\ i -> missed i / fromIntegral (min nLabels i))
  where missed i = fromIntegral $ length (takeWhile (< i) hitIndices)


evaluateLine :: Labels -> Labels -> Maybe Statistics
evaluateLine []   _     = Nothing
evaluateLine _    []    = Nothing
evaluateLine lbls preds = Just Statistics {
    auc       = calculateAUC nLabels nPreds hitIndices
  , cover     = fromIntegral (length lblsInPreds) / nLabelsN
  , rank      = calculateRank nLabels nPreds hitIndices
  , precision = length predsInLbls
  , recall    = recall'
  , coverI    = calculateCoverI nLabels hitIndices indices
  , recallI   = map (\ i -> if i >= recall' then 1 else 0) indices
  , lblRanks  = map (\ l -> (l, IM.lookup l preds100M)) lbls
  }
  where
    nLabels   = length lbls
    nPreds    = length preds
    nLabelsN  = fromIntegral nLabels
    preds100  = take 100 preds
    preds100M = IM.fromList $ zip preds100 [1 ..]
    lblsS     = IS.fromList lbls

    lblsInPreds = filter (`IM.member` preds100M) lbls
    predsInLbls = filter (`IS.member` lblsS) preds100
    hitIndices  = findIndices (`IS.member` lblsS) preds

    indices = [1 .. 200]
    --indices = [2^x | x <- [0 .. 11]]

    recall'  = calculateRecall nLabels nPreds hitIndices

liftSnd :: Functor m => (a, m b) -> m (a, b)
liftSnd (x, y) = fmap (\ y' -> (x, y')) y

evaluateLines :: Seq Labels -> Seq Labels -> [(LabelID, Statistics)]
evaluateLines lbls =
  mapMaybe (\ (l, p) -> liftSnd (l, evaluateLine (Seq.index lbls l) p)) .
  zip [0 ..] . toList

sumLists :: Num a => [[a]] -> [a]
sumLists = foldl1 (zipWith (+))

combineStatistics :: [(LabelID, Statistics)] -> FileStatistics
combineStatistics labelStats = FileStatistics {
    aucF       = sum (map auc       stats) / n
  , coverF     = sum (map cover     stats) / n
  , rankF      = sum (map rank      stats) / n
  , precisionF = fromIntegral (sum (map precision stats)) / n
  , recallF    = fromIntegral (sum (map recall    stats)) / n
  , coverIF    = map (100 / n *) $ sumLists $ map coverI stats
  , recallIF   = sumLists $ map recallI stats
  , lblRanksF  = map (second lblRanks) labelStats
  }
  where stats = map snd labelStats
        n     = fromIntegral $ length stats

statisticsList :: FilePath -> FileStatistics -> [String]
statisticsList fp stats =
  fp : zipWith (\ f d -> printf ("%." ++ show (d :: Int) ++ "f") $ f stats)
  [coverF, precisionF, recallF, aucF, rankF]
  [     3,          3,       2,    4,     2]

showLblRanks :: LabelMap -> FileStatistics -> [String]
showLblRanks lblMap =
  map (\ (l, r) -> unlines $ (showLabel l ++ ":") : showRanks r) . lblRanksF
  where
    showLabel = BC.unpack . IBM.findElement lblMap
    showRanks = map (uncurry showRank) . sortBy (comparing snd)
    showRank lbl rnk = "  " ++
      maybe "N/A" (printf "%3d") rnk ++ " -> " ++ showLabel lbl


data Options = Options {
  optCSV     :: Bool
, optCoverI  :: Bool
, optRecallI :: Bool
, optRanks   :: Bool
}

defaultOptions :: Options
defaultOptions = Options {
  optCSV     = False
, optCoverI  = False
, optRecallI = False
, optRanks   = False
}

options :: [OptDescr (Options -> Options)]
options =
  [
    Option "t" []
      (NoArg (\ opts -> opts {optCSV = True})) "output tables in CSV format"
  , Option "c" []
      (NoArg (\ opts -> opts {optCoverI  = True})) "print Cover-i"
  , Option "r" []
      (NoArg (\ opts -> opts {optRecallI = True})) "print Recall-i"
  , Option "p" []
      (NoArg (\ opts -> opts {optRanks = True})) "print prediction ranks"
  ]

header :: String -> String
header progName = intercalate "\n"
  [ "Usage: " ++ progName ++ " <seq> <deps> [OPTS] [<preds1> [... [<preds_n>]]]"
  , ""
  , "Mandatory arguments:"
  , "  <seq>    file containing the order of theorems"
  , "  <deps>   file containing the dependencies of theorems"
  , ""
  , "Optional arguments:"
  ]

showUsage :: IO ()
showUsage = do
  progName <- getProgName
  hPutStrLn stderr $ usageInfo (header progName) options

main :: IO ()
main = do
  args <- getArgs
  (seqFP, labelsFP, pFPs, opts) <- case getOpt Permute options args of
    (o, sF : lF : pFs, []) ->
      return (sF, lF, pFs, foldl (flip id) defaultOptions o)
    _ -> showUsage >> exitFailure

  lblMap <- I.loadSequenceFile seqFP
  lbls   <- I.loadLabelsFile lblMap labelsFP
  preds  <- if null pFPs
           then liftM (pure . I.loadLabels lblMap . BC.lines) BC.getContents
           else mapM (I.loadLabelsFile lblMap) pFPs

  let statistics = parMap rdeepseq (combineStatistics . evaluateLines lbls) preds
      predFPs = if null pFPs then ["<stdin>"] else pFPs

  let descriptors = ["", "Cov", "Prec", "Rec", "AUC", "Rank"]
      statisticsOpts = TableOptions {
        separator = if optCSV opts then ";" else " "
      , cellWidths = maximum (map length predFPs) : repeat 7
      , alignments = AlignLeft : repeat AlignRight
      }
      graphOpts = statisticsOpts {
        cellWidths = map length predFPs
      , alignments = repeat AlignRight
      }

      stats = descriptors : zipWith statisticsList predFPs statistics
      covIs = zipWith (\ fp s -> fp : show2fs  (coverIF  s)) predFPs statistics
      recIs = zipWith (\ fp s -> fp : map show (recallIF s)) predFPs statistics
      lRnks = zipWith (\ fp s -> fp : showLblRanks lblMap s) predFPs statistics

  let showStats = not $ optCoverI opts || optRecallI opts || optRanks opts

  when showStats $ putStr $ renderTable statisticsOpts stats
  when (optCoverI  opts) $ putStr $ renderTable graphOpts $ transpose covIs
  when (optRecallI opts) $ putStr $ renderTable graphOpts $ transpose recIs
  when (optRanks   opts) $ mapM_ (putStrLn . unlines) lRnks

  where show2fs = map (printf "%.2f")
